@@ -13,7 +13,7 @@
 #include "php_http_api.h"
 #include "php_http_client.h"
 
-#include <ext/spl/spl_observer.h>
+#include "ext/spl/spl_observer.h"
 
 /*
  * array of name => php_http_client_driver_t*
@@ -74,7 +74,8 @@ void php_http_client_options_set_subr(zval *instance, char *key, size_t len, zva
 		zval old_opts_tmp, *old_opts, new_opts, *entry = NULL;
 
 		array_init(&new_opts);
-		old_opts = zend_read_property(this_ce, instance, ZEND_STRL("options"), 0, &old_opts_tmp);
+		old_opts = zend_read_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), 0, &old_opts_tmp);
+
 		if (Z_TYPE_P(old_opts) == IS_ARRAY) {
 			array_copy(Z_ARRVAL_P(old_opts), Z_ARRVAL(new_opts));
 		}
@@ -88,6 +89,7 @@ void php_http_client_options_set_subr(zval *instance, char *key, size_t len, zva
 			}
 		} else if (opts && zend_hash_num_elements(Z_ARRVAL_P(opts))) {
 			if ((entry = zend_symtable_str_find(Z_ARRVAL(new_opts), key, len))) {
+				SEPARATE_ZVAL(entry);
 				array_join(Z_ARRVAL_P(opts), Z_ARRVAL_P(entry), 0, 0);
 			} else {
 				Z_ADDREF_P(opts);
@@ -95,7 +97,7 @@ void php_http_client_options_set_subr(zval *instance, char *key, size_t len, zva
 			}
 		}
 
-		zend_update_property(this_ce, instance, ZEND_STRL("options"), &new_opts);
+		zend_update_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), &new_opts);
 		zval_ptr_dtor(&new_opts);
 	}
 }
@@ -110,7 +112,7 @@ void php_http_client_options_set(zval *instance, zval *opts)
 	array_init(&new_opts);
 
 	if (!opts || !zend_hash_num_elements(Z_ARRVAL_P(opts))) {
-		zend_update_property(this_ce, instance, ZEND_STRL("options"), &new_opts);
+		zend_update_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), &new_opts);
 		zval_ptr_dtor(&new_opts);
 	} else {
 		zval old_opts_tmp, *old_opts, add_opts, *opt;
@@ -123,9 +125,9 @@ void php_http_client_options_set(zval *instance, zval *opts)
 				if (Z_TYPE_P(opt) == IS_ARRAY && (zend_string_equals_literal(key.key, "ssl") || zend_string_equals_literal(key.key, "cookies"))) {
 					php_http_client_options_set_subr(instance, key.key->val, key.key->len, opt, 0);
 				} else if (is_client && (zend_string_equals_literal(key.key, "recordHistory") || zend_string_equals_literal(key.key, "responseMessageClass"))) {
-					zend_update_property(this_ce, instance, key.key->val, key.key->len, opt);
+					zend_update_property(this_ce, Z_OBJ_P(instance), key.key->val, key.key->len, opt);
 				} else if (Z_TYPE_P(opt) == IS_NULL) {
-					old_opts = zend_read_property(this_ce, instance, ZEND_STRL("options"), 0, &old_opts_tmp);
+					old_opts = zend_read_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), 0, &old_opts_tmp);
 					if (Z_TYPE_P(old_opts) == IS_ARRAY) {
 						zend_symtable_del(Z_ARRVAL_P(old_opts), key.key);
 					}
@@ -137,12 +139,12 @@ void php_http_client_options_set(zval *instance, zval *opts)
 		}
 		ZEND_HASH_FOREACH_END();
 
-		old_opts = zend_read_property(this_ce, instance, ZEND_STRL("options"), 0, &old_opts_tmp);
+		old_opts = zend_read_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), 0, &old_opts_tmp);
 		if (Z_TYPE_P(old_opts) == IS_ARRAY) {
 			array_copy(Z_ARRVAL_P(old_opts), Z_ARRVAL(new_opts));
 		}
 		array_join(Z_ARRVAL(add_opts), Z_ARRVAL(new_opts), 0, 0);
-		zend_update_property(this_ce, instance, ZEND_STRL("options"), &new_opts);
+		zend_update_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), &new_opts);
 		zval_ptr_dtor(&new_opts);
 		zval_ptr_dtor(&add_opts);
 	}
@@ -151,7 +153,7 @@ void php_http_client_options_set(zval *instance, zval *opts)
 void php_http_client_options_get_subr(zval *instance, char *key, size_t len, zval *return_value)
 {
 	zend_class_entry *this_ce = Z_OBJCE_P(instance);
-	zval *options, opts_tmp, *opts = zend_read_property(this_ce, instance, ZEND_STRL("options"), 0, &opts_tmp);
+	zval *options, opts_tmp, *opts = zend_read_property(this_ce, Z_OBJ_P(instance), ZEND_STRL("options"), 0, &opts_tmp);
 
 	if ((Z_TYPE_P(opts) == IS_ARRAY) && (options = zend_symtable_str_find(Z_ARRVAL_P(opts), key, len))) {
 		RETVAL_ZVAL(options, 1, 0);
@@ -211,6 +213,8 @@ void php_http_client_dtor(php_http_client_t *h)
 	if (h->ops->dtor) {
 		h->ops->dtor(h);
 	}
+	h->callback.debug.func = NULL;
+	h->callback.debug.arg = NULL;
 
 	php_resource_factory_free(&h->rf);
 }
@@ -250,6 +254,20 @@ ZEND_RESULT_CODE php_http_client_dequeue(php_http_client_t *h, php_http_message_
 	return FAILURE;
 }
 
+ZEND_RESULT_CODE php_http_client_requeue(php_http_client_t *h, php_http_message_t *request)
+{
+	if (h->ops->dequeue) {
+		php_http_client_enqueue_t *enqueue = php_http_client_enqueued(h, request, NULL);
+
+		if (!enqueue) {
+			php_error_docref(NULL, E_WARNING, "Failed to requeue request; request not in queue");
+			return FAILURE;
+		}
+		return h->ops->requeue(h, enqueue);
+	}
+	return FAILURE;
+}
+
 php_http_client_enqueue_t *php_http_client_enqueued(php_http_client_t *h, void *compare_arg, php_http_client_enqueue_cmp_func_t compare_func)
 {
 	zend_llist_element *el = NULL;
@@ -330,7 +348,14 @@ void php_http_client_object_free(zend_object *object)
 {
 	php_http_client_object_t *o = PHP_HTTP_OBJ(object, NULL);
 
+	PTR_FREE(o->gc);
+
 	php_http_client_free(&o->client);
+	if (o->debug.fci.size > 0) {
+		zend_fcall_info_args_clear(&o->debug.fci, 1);
+		zval_ptr_dtor(&o->debug.fci.function_name);
+		o->debug.fci.size = 0;
+	}
 	php_http_object_method_dtor(&o->notify);
 	php_http_object_method_free(&o->update);
 	zend_object_std_dtor(object);
@@ -356,9 +381,57 @@ zend_object *php_http_client_object_new(zend_class_entry *ce)
 	return &php_http_client_object_new_ex(ce, NULL)->zo;
 }
 
+static HashTable *php_http_client_object_get_gc(zend_object *object, zval **table, int *n)
+{
+	php_http_client_object_t *obj = PHP_HTTP_OBJ(object, NULL);
+	zend_llist_element *el = NULL;
+	HashTable *props = object->handlers->get_properties(object);
+	uint32_t count = zend_hash_num_elements(props) + zend_llist_count(&obj->client->responses) + zend_llist_count(&obj->client->requests) + 2;
+	zval *val;
+
+	*n = 0;
+	*table = obj->gc = erealloc(obj->gc, sizeof(zval) * count);
+
+#if PHP_HTTP_HAVE_LIBCURL
+	if (obj->client->ops == php_http_client_curl_get_ops()) {
+		php_http_client_curl_t *curl = obj->client->ctx;
+
+		if (curl->ev_ops == php_http_client_curl_user_ops_get()) {
+			php_http_client_curl_user_context_t *ctx = curl->ev_ctx;
+
+			ZVAL_COPY_VALUE(&obj->gc[(*n)++], &ctx->user);
+		}
+	}
+#endif
+
+	if (obj->debug.fci.size > 0) {
+		ZVAL_COPY_VALUE(&obj->gc[(*n)++], &obj->debug.fci.function_name);
+	}
+
+	for (el = obj->client->responses.head; el; el = el->next) {
+		php_http_message_object_t *response_obj = *(php_http_message_object_t **) el->data;
+		ZVAL_OBJ(&obj->gc[(*n)++], &response_obj->zo);
+	}
+
+	for (el = obj->client->requests.head; el; el = el->next) {
+		php_http_client_enqueue_t *q = (php_http_client_enqueue_t *) el->data;
+		if (q->request_obj) {
+			ZVAL_OBJ(&obj->gc[(*n)++], &q->request_obj->zo);
+		}
+	}
+
+	ZEND_HASH_FOREACH_VAL(props, val)
+	{
+		ZVAL_COPY_VALUE(&obj->gc[(*n)++], val);
+	}
+	ZEND_HASH_FOREACH_END();
+
+	return NULL;
+}
+
 static void handle_history(zval *zclient, php_http_message_t *request, php_http_message_t *response)
 {
-	zval new_hist, old_hist_tmp, *old_hist = zend_read_property(php_http_client_class_entry, zclient, ZEND_STRL("history"), 0, &old_hist_tmp);
+	zval new_hist, old_hist_tmp, *old_hist = zend_read_property(php_http_client_class_entry, Z_OBJ_P(zclient), ZEND_STRL("history"), 0, &old_hist_tmp);
 	php_http_message_t *req_copy = php_http_message_copy(request, NULL);
 	php_http_message_t *res_copy = php_http_message_copy(response, NULL);
 	php_http_message_t *zipped = php_http_message_zip(res_copy, req_copy);
@@ -370,7 +443,7 @@ static void handle_history(zval *zclient, php_http_message_t *request, php_http_
 		php_http_message_object_prepend(&new_hist, old_hist, 1);
 	}
 
-	zend_update_property(php_http_client_class_entry, zclient, ZEND_STRL("history"), &new_hist);
+	zend_update_property(php_http_client_class_entry, Z_OBJ_P(zclient), ZEND_STRL("history"), &new_hist);
 	zval_ptr_dtor(&new_hist);
 }
 
@@ -391,7 +464,7 @@ static ZEND_RESULT_CODE handle_response(void *arg, php_http_client_t *client, ph
 		/* ensure the message is of type response (could be uninitialized in case of early error, like DNS) */
 		php_http_message_set_type(msg, PHP_HTTP_RESPONSE);
 
-		if (zend_is_true(zend_read_property(php_http_client_class_entry, &zclient, ZEND_STRL("recordHistory"), 0, &rec_hist_tmp))) {
+		if (zend_is_true(zend_read_property(php_http_client_class_entry, Z_OBJ(zclient), ZEND_STRL("recordHistory"), 0, &rec_hist_tmp))) {
 			handle_history(&zclient, e->request, *response);
 		}
 
@@ -400,7 +473,7 @@ static ZEND_RESULT_CODE handle_response(void *arg, php_http_client_t *client, ph
 		*response = NULL;
 
 		msg_obj = php_http_message_object_new_ex(php_http_get_client_response_class_entry(), msg);
-		ZVAL_OBJ(&zresponse, &msg_obj->zo);
+		ZVAL_OBJECT(&zresponse, &msg_obj->zo, 1);
 		ZVAL_OBJECT(&zrequest, &((php_http_message_object_t *) e->opaque)->zo, 1);
 
 		php_http_message_object_prepend(&zresponse, &zrequest, 1);
@@ -408,10 +481,9 @@ static ZEND_RESULT_CODE handle_response(void *arg, php_http_client_t *client, ph
 		object_init(&info);
 		info_ht = HASH_OF(&info);
 		php_http_client_getopt(client, PHP_HTTP_CLIENT_OPT_TRANSFER_INFO, e->request, &info_ht);
-		zend_update_property(php_http_get_client_response_class_entry(), &zresponse, ZEND_STRL("transferInfo"), &info);
+		zend_update_property(php_http_get_client_response_class_entry(), Z_OBJ(zresponse), ZEND_STRL("transferInfo"), &info);
 		zval_ptr_dtor(&info);
 
-		Z_ADDREF(zresponse);
 		zend_llist_add_element(&client->responses, &msg_obj);
 
 		if (e->closure.fci.size) {
@@ -421,7 +493,9 @@ static ZEND_RESULT_CODE handle_response(void *arg, php_http_client_t *client, ph
 			ZVAL_UNDEF(&retval);
 			zend_fcall_info_argn(&e->closure.fci, 1, &zresponse);
 			zend_replace_error_handling(EH_NORMAL, NULL, &zeh);
+			++client->callback.depth;
 			zend_fcall_info_call(&e->closure.fci, &e->closure.fcc, &retval, NULL);
+			--client->callback.depth;
 			zend_restore_error_handling(&zeh);
 			zend_fcall_info_argn(&e->closure.fci, 0);
 
@@ -466,7 +540,9 @@ static void handle_progress(void *arg, php_http_client_t *client, php_http_clien
 	add_property_double(&args[1], "ulnow", progress->ul.now);
 
 	zend_replace_error_handling(EH_NORMAL, NULL, &zeh);
+	++client->callback.depth;
 	php_http_object_method_call(&client_obj->notify, &zclient, NULL, 2, args);
+	--client->callback.depth;
 	zend_restore_error_handling(&zeh);
 
 	zval_ptr_dtor(&zclient);
@@ -474,11 +550,36 @@ static void handle_progress(void *arg, php_http_client_t *client, php_http_clien
 	zval_ptr_dtor(&args[1]);
 }
 
+static void handle_debug(void *arg, php_http_client_t *client, php_http_client_enqueue_t *e, unsigned type, const char *data, size_t size)
+{
+	zval ztype, zdata, zreq, zclient;
+	php_http_client_object_t *client_obj = arg;
+	zend_error_handling zeh;
+
+	ZVAL_OBJECT(&zclient, &client_obj->zo, 1);
+	ZVAL_OBJECT(&zreq, &((php_http_message_object_t *) e->opaque)->zo, 1);
+	ZVAL_LONG(&ztype, type);
+	ZVAL_STRINGL(&zdata, data, size);
+
+	zend_replace_error_handling(EH_NORMAL, NULL, &zeh);
+	zend_fcall_info_argn(&client_obj->debug.fci, 4, &zclient, &zreq, &ztype, &zdata);
+	++client->callback.depth;
+	zend_fcall_info_call(&client_obj->debug.fci, &client_obj->debug.fcc, NULL, NULL);
+	--client->callback.depth;
+	zend_fcall_info_args_clear(&client_obj->debug.fci, 0);
+	zend_restore_error_handling(&zeh);
+
+	zval_ptr_dtor(&zclient);
+	zval_ptr_dtor(&zreq);
+	zval_ptr_dtor(&ztype);
+	zval_ptr_dtor(&zdata);
+}
+
 static void response_dtor(void *data)
 {
 	php_http_message_object_t *msg_obj = *(php_http_message_object_t **) data;
 
-	zend_objects_store_del(&msg_obj->zo);
+	zend_object_release(&msg_obj->zo);
 }
 
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_construct, 0, 0, 0)
@@ -496,7 +597,7 @@ static PHP_METHOD(HttpClient, __construct)
 		php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "|S!S!", &driver_name, &persistent_handle_name), invalid_arg, return);
 
 		if (!zend_hash_num_elements(&php_http_client_drivers)) {
-			php_http_throw(unexpected_val, "No http\\Client drivers available", NULL);
+			php_http_throw(unexpected_val, "No http\\Client drivers available");
 			return;
 		}
 		if (!(driver = php_http_client_driver_get(driver_name))) {
@@ -505,7 +606,7 @@ static PHP_METHOD(HttpClient, __construct)
 		}
 
 		object_init_ex(&os, spl_ce_SplObjectStorage);
-		zend_update_property(php_http_client_class_entry, getThis(), ZEND_STRL("observers"), &os);
+		zend_update_property(php_http_client_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("observers"), &os);
 		zval_ptr_dtor(&os);
 
 		if (persistent_handle_name) {
@@ -545,25 +646,28 @@ static PHP_METHOD(HttpClient, reset)
 	RETVAL_ZVAL(getThis(), 1, 0);
 }
 
-static HashTable *combined_options(zval *client, zval *request)
+static HashTable *combined_options(HashTable *options, zval *client, zval *request)
 {
-	HashTable *options;
 	unsigned num_options = 0;
-	zval z_roptions, z_options_tmp, *z_coptions = zend_read_property(php_http_client_class_entry, client, ZEND_STRL("options"), 0, &z_options_tmp);
+	zval z_roptions, z_options_tmp, *z_coptions = zend_read_property(php_http_client_class_entry, Z_OBJ_P(client), ZEND_STRL("options"), 0, &z_options_tmp);
 
 	if (Z_TYPE_P(z_coptions) == IS_ARRAY) {
 		num_options = zend_hash_num_elements(Z_ARRVAL_P(z_coptions));
 	}
 	ZVAL_UNDEF(&z_roptions);
-	zend_call_method_with_0_params(request, NULL, NULL, "getOptions", &z_roptions);
+	zend_call_method_with_0_params(Z_OBJ_P(request), NULL, NULL, "getOptions", &z_roptions);
 	if (Z_TYPE(z_roptions) == IS_ARRAY) {
 		unsigned num = zend_hash_num_elements(Z_ARRVAL(z_roptions));
 		if (num > num_options) {
 			num_options = num;
 		}
 	}
-	ALLOC_HASHTABLE(options);
-	ZEND_INIT_SYMTABLE_EX(options, num_options, 0);
+	if (options) {
+		zend_hash_clean(options);
+	} else {
+		ALLOC_HASHTABLE(options);
+		ZEND_INIT_SYMTABLE_EX(options, num_options, 0);
+	}
 	if (Z_TYPE_P(z_coptions) == IS_ARRAY) {
 		array_copy(Z_ARRVAL_P(z_coptions), options);
 	}
@@ -579,14 +683,14 @@ static void msg_queue_dtor(php_http_client_enqueue_t *e)
 {
 	php_http_message_object_t *msg_obj = e->opaque;
 
-	zend_objects_store_del(&msg_obj->zo);
+	zend_object_release(&msg_obj->zo);
 	zend_hash_destroy(e->options);
 	FREE_HASHTABLE(e->options);
 
 	if (e->closure.fci.size) {
 		zval_ptr_dtor(&e->closure.fci.function_name);
 		if (e->closure.fci.object) {
-			zend_objects_store_del(e->closure.fci.object);
+			zend_object_release(e->closure.fci.object);
 		}
 	}
 }
@@ -602,7 +706,7 @@ static PHP_METHOD(HttpClient, enqueue)
 	zend_fcall_info_cache fcc = empty_fcall_info_cache;
 	php_http_client_object_t *obj;
 	php_http_message_object_t *msg_obj;
-	php_http_client_enqueue_t q;
+	php_http_client_enqueue_t q = {0};
 
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "O|f", &request, php_http_get_client_request_class_entry(), &fci, &fcc), invalid_arg, return);
 
@@ -610,12 +714,23 @@ static PHP_METHOD(HttpClient, enqueue)
 	msg_obj = PHP_HTTP_OBJ(NULL, request);
 
 	if (php_http_client_enqueued(obj->client, msg_obj->message, NULL)) {
-		php_http_throw(bad_method_call, "Failed to enqueue request; request already in queue", NULL);
+		php_http_throw(bad_method_call, "Failed to enqueue request; request already in queue");
 		return;
 	}
 
+	/* set early for progress callback */
+	q.opaque = msg_obj;
+
+	if (obj->client->callback.progress.func) {
+		php_http_client_progress_state_t progress = {0};
+
+		progress.info = "prepare";
+		obj->client->callback.progress.func(obj->client->callback.progress.arg, obj->client, &q, &progress);
+	}
+
+	Z_ADDREF_P(request);
 	q.request = msg_obj->message;
-	q.options = combined_options(getThis(), request);
+	q.options = combined_options(NULL, getThis(), request);
 	q.dtor = msg_queue_dtor;
 	q.opaque = msg_obj;
 	q.closure.fci = fci;
@@ -624,12 +739,10 @@ static PHP_METHOD(HttpClient, enqueue)
 	if (fci.size) {
 		Z_TRY_ADDREF(fci.function_name);
 		if (fci.object) {
-			++GC_REFCOUNT(fci.object);
+			GC_ADDREF(fci.object);
 		}
 	}
 
-	Z_ADDREF_P(request);
-
 	php_http_expect(SUCCESS == php_http_client_enqueue(obj->client, &q), runtime,
 			msg_queue_dtor(&q);
 			return;
@@ -653,7 +766,7 @@ static PHP_METHOD(HttpClient, dequeue)
 	msg_obj = PHP_HTTP_OBJ(NULL, request);
 
 	if (!php_http_client_enqueued(obj->client, msg_obj->message, NULL)) {
-		php_http_throw(bad_method_call, "Failed to dequeue request; request not in queue", NULL);
+		php_http_throw(bad_method_call, "Failed to dequeue request; request not in queue");
 		return;
 	}
 
@@ -673,19 +786,35 @@ static PHP_METHOD(HttpClient, requeue)
 	zend_fcall_info_cache fcc = empty_fcall_info_cache;
 	php_http_client_object_t *obj;
 	php_http_message_object_t *msg_obj;
-	php_http_client_enqueue_t q;
+	php_http_client_enqueue_t q, *e;
 
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "O|f", &request, php_http_get_client_request_class_entry(), &fci, &fcc), invalid_arg, return);
 
 	obj = PHP_HTTP_OBJ(NULL, getThis());
 	msg_obj = PHP_HTTP_OBJ(NULL, request);
 
-	if (php_http_client_enqueued(obj->client, msg_obj->message, NULL)) {
-		php_http_expect(SUCCESS == php_http_client_dequeue(obj->client, msg_obj->message), runtime, return);
+	if ((e = php_http_client_enqueued(obj->client, msg_obj->message, NULL))) {
+		combined_options(e->options, getThis(), request);
+		php_http_expect(SUCCESS == php_http_client_requeue(obj->client, msg_obj->message), runtime, return);
+		if (fci.size) {
+			if (e->closure.fci.size) {
+				zval_ptr_dtor(&e->closure.fci.function_name);
+				if (e->closure.fci.object) {
+					zend_object_release(e->closure.fci.object);
+				}
+			}
+			Z_TRY_ADDREF(fci.function_name);
+			if (fci.object) {
+				GC_ADDREF(fci.object);
+			}
+			e->closure.fci = fci;
+			e->closure.fcc = fcc;
+		}
+		RETURN_ZVAL(getThis(), 1, 0);
 	}
 
 	q.request = msg_obj->message;
-	q.options = combined_options(getThis(), request);
+	q.options = combined_options(NULL, getThis(), request);
 	q.dtor = msg_queue_dtor;
 	q.opaque = msg_obj;
 	q.closure.fci = fci;
@@ -694,7 +823,7 @@ static PHP_METHOD(HttpClient, requeue)
 	if (fci.size) {
 		Z_TRY_ADDREF(fci.function_name);
 		if (fci.object) {
-			++GC_REFCOUNT(fci.object);
+			GC_ADDREF(fci.object);
 		}
 	}
 
@@ -708,7 +837,7 @@ static PHP_METHOD(HttpClient, requeue)
 	RETVAL_ZVAL(getThis(), 1, 0);
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_count, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpClient_count, 0, 0, IS_LONG, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpClient, count)
 {
@@ -747,7 +876,7 @@ static PHP_METHOD(HttpClient, getResponse)
 		}
 
 		/* not found for the request! */
-		php_http_throw(unexpected_val, "Could not find response for the request", NULL);
+		php_http_throw(unexpected_val, "Could not find response for the request");
 		return;
 	}
 
@@ -771,7 +900,7 @@ static PHP_METHOD(HttpClient, getHistory)
 
 	php_http_expect(SUCCESS == zend_parse_parameters_none(), invalid_arg, return);
 
-	zhistory = zend_read_property(php_http_client_class_entry, getThis(), ZEND_STRL("history"), 0, &zhistory_tmp);
+	zhistory = zend_read_property(php_http_client_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("history"), 0, &zhistory_tmp);
 	RETVAL_ZVAL(zhistory, 1, 0);
 }
 
@@ -888,8 +1017,13 @@ static int notify(zend_object_iterator *iter, void *puser)
 	return ZEND_HASH_APPLY_STOP;
 }
 
+#if PHP_VERSION_ID >= 80100
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpClient_notify, 0, 0, IS_VOID, 0)
+#else
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_notify, 0, 0, 0)
+#endif
 	ZEND_ARG_OBJ_INFO(0, request, http\\Client\\Request, 1)
+	ZEND_ARG_INFO(0, progress)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpClient, notify)
 {
@@ -900,10 +1034,10 @@ static PHP_METHOD(HttpClient, notify)
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "|O!o!", &request, php_http_get_client_request_class_entry(), &zprogress), invalid_arg, return);
 
 	client_obj = PHP_HTTP_OBJ(NULL, getThis());
-	observers = zend_read_property(php_http_client_class_entry, getThis(), ZEND_STRL("observers"), 0, &observers_tmp);
+	observers = zend_read_property(php_http_client_class_entry, &client_obj->zo, ZEND_STRL("observers"), 0, &observers_tmp);
 
 	if (Z_TYPE_P(observers) != IS_OBJECT) {
-		php_http_throw(unexpected_val, "Observer storage is corrupted", NULL);
+		php_http_throw(unexpected_val, "Observer storage is corrupted");
 		return;
 	}
 
@@ -931,11 +1065,16 @@ static PHP_METHOD(HttpClient, notify)
 			zval_ptr_dtor(zprogress);
 		}
 	}
-
+#if PHP_VERSION_ID < 80100
 	RETVAL_ZVAL(getThis(), 1, 0);
+#endif
 }
 
+#if PHP_VERSION_ID >= 80100
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpClient_attach, 0, 1, IS_VOID, 0)
+#else
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_attach, 0, 0, 1)
+#endif
 	ZEND_ARG_OBJ_INFO(0, observer, SplObserver, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpClient, attach)
@@ -946,10 +1085,10 @@ static PHP_METHOD(HttpClient, attach)
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "O", &observer, spl_ce_SplObserver), invalid_arg, return);
 
 	client_obj = PHP_HTTP_OBJ(NULL, getThis());
-	observers = zend_read_property(php_http_client_class_entry, getThis(), ZEND_STRL("observers"), 0, &observers_tmp);
+	observers = zend_read_property(php_http_client_class_entry, &client_obj->zo, ZEND_STRL("observers"), 0, &observers_tmp);
 
 	if (Z_TYPE_P(observers) != IS_OBJECT) {
-		php_http_throw(unexpected_val, "Observer storage is corrupted", NULL);
+		php_http_throw(unexpected_val, "Observer storage is corrupted");
 		return;
 	}
 
@@ -958,13 +1097,18 @@ static PHP_METHOD(HttpClient, attach)
 	}
 
 	ZVAL_UNDEF(&retval);
-	zend_call_method_with_1_params(observers, NULL, NULL, "attach", &retval, observer);
+	zend_call_method_with_1_params(Z_OBJ_P(observers), NULL, NULL, "attach", &retval, observer);
 	zval_ptr_dtor(&retval);
-
+#if PHP_VERSION_ID < 80100
 	RETVAL_ZVAL(getThis(), 1, 0);
+#endif
 }
 
+#if PHP_VERSION_ID >= 80100
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpClient_detach, 0, 1, IS_VOID, 0)
+#else
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_detach, 0, 0, 1)
+#endif
 	ZEND_ARG_OBJ_INFO(0, observer, SplObserver, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpClient, detach)
@@ -973,18 +1117,19 @@ static PHP_METHOD(HttpClient, detach)
 
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "O", &observer, spl_ce_SplObserver), invalid_arg, return);
 
-	observers = zend_read_property(php_http_client_class_entry, getThis(), ZEND_STRL("observers"), 0, &observers_tmp);
+	observers = zend_read_property(php_http_client_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("observers"), 0, &observers_tmp);
 
 	if (Z_TYPE_P(observers) != IS_OBJECT) {
-		php_http_throw(unexpected_val, "Observer storage is corrupted", NULL);
+		php_http_throw(unexpected_val, "Observer storage is corrupted");
 		return;
 	}
 
 	ZVAL_UNDEF(&retval);
-	zend_call_method_with_1_params(observers, NULL, NULL, "detach", &retval, observer);
+	zend_call_method_with_1_params(Z_OBJ_P(observers), NULL, NULL, "detach", &retval, observer);
 	zval_ptr_dtor(&retval);
-
+#if PHP_VERSION_ID < 80100
 	RETVAL_ZVAL(getThis(), 1, 0);
+#endif
 }
 
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_getObservers, 0, 0, 0)
@@ -995,10 +1140,10 @@ static PHP_METHOD(HttpClient, getObservers)
 
 	php_http_expect(SUCCESS == zend_parse_parameters_none(), invalid_arg, return);
 
-	observers = zend_read_property(php_http_client_class_entry, getThis(), ZEND_STRL("observers"), 0, &observers_tmp);
+	observers = zend_read_property(php_http_client_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("observers"), 0, &observers_tmp);
 
 	if (Z_TYPE_P(observers) != IS_OBJECT) {
-		php_http_throw(unexpected_val, "Observer storage is corrupted", NULL);
+		php_http_throw(unexpected_val, "Observer storage is corrupted");
 		return;
 	}
 
@@ -1071,7 +1216,7 @@ ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpClient, getOptions)
 {
 	if (SUCCESS == zend_parse_parameters_none()) {
-		zval options_tmp, *options = zend_read_property(php_http_client_class_entry, getThis(), ZEND_STRL("options"), 0, &options_tmp);
+		zval options_tmp, *options = zend_read_property(php_http_client_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("options"), 0, &options_tmp);
 		RETVAL_ZVAL(options, 1, 0);
 	}
 }
@@ -1184,8 +1329,41 @@ static PHP_METHOD(HttpClient, getAvailableConfiguration)
 	}
 }
 
+ZEND_BEGIN_ARG_INFO_EX(ai_HttpClient_setDebug, 0, 0, 1)
+	/* using IS_CALLABLE type hint would create a forwards compatibility break */
+	ZEND_ARG_INFO(0, callback)
+ZEND_END_ARG_INFO();
+static PHP_METHOD(HttpClient, setDebug)
+{
+	zend_fcall_info fci;
+	zend_fcall_info_cache fcc;
+	php_http_client_object_t *client_obj;
+
+	fci.size = 0;
+	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "|f", &fci, &fcc), invalid_arg, return);
+
+	client_obj = PHP_HTTP_OBJ(NULL, getThis());
+
+	if (client_obj->debug.fci.size > 0) {
+		zval_ptr_dtor(&client_obj->debug.fci.function_name);
+		client_obj->debug.fci.size = 0;
+	}
+	if (fci.size > 0) {
+		memcpy(&client_obj->debug.fci, &fci, sizeof(fci));
+		memcpy(&client_obj->debug.fcc, &fcc, sizeof(fcc));
+		Z_ADDREF_P(&fci.function_name);
+		client_obj->client->callback.debug.func = handle_debug;
+		client_obj->client->callback.debug.arg = client_obj;
+	} else {
+		client_obj->client->callback.debug.func = NULL;
+		client_obj->client->callback.debug.arg = NULL;
+	}
+
+	RETVAL_ZVAL(getThis(), 1, 0);
+}
+
 static zend_function_entry php_http_client_methods[] = {
-	PHP_ME(HttpClient, __construct,          ai_HttpClient_construct,            ZEND_ACC_PUBLIC|ZEND_ACC_CTOR)
+	PHP_ME(HttpClient, __construct,          ai_HttpClient_construct,            ZEND_ACC_PUBLIC)
 	PHP_ME(HttpClient, reset,                ai_HttpClient_reset,                ZEND_ACC_PUBLIC)
 	PHP_ME(HttpClient, enqueue,              ai_HttpClient_enqueue,              ZEND_ACC_PUBLIC)
 	PHP_ME(HttpClient, dequeue,              ai_HttpClient_dequeue,              ZEND_ACC_PUBLIC)
@@ -1216,6 +1394,7 @@ static zend_function_entry php_http_client_methods[] = {
 	PHP_ME(HttpClient, getAvailableDrivers,  ai_HttpClient_getAvailableDrivers,  ZEND_ACC_PUBLIC|ZEND_ACC_STATIC)
 	PHP_ME(HttpClient, getAvailableOptions,  ai_HttpClient_getAvailableOptions,  ZEND_ACC_PUBLIC)
 	PHP_ME(HttpClient, getAvailableConfiguration, ai_HttpClient_getAvailableConfiguration, ZEND_ACC_PUBLIC)
+	PHP_ME(HttpClient, setDebug,             ai_HttpClient_setDebug,             ZEND_ACC_PUBLIC)
 	EMPTY_FUNCTION_ENTRY
 };
 
@@ -1226,16 +1405,24 @@ PHP_MINIT_FUNCTION(http_client)
 	INIT_NS_CLASS_ENTRY(ce, "http", "Client", php_http_client_methods);
 	php_http_client_class_entry = zend_register_internal_class_ex(&ce, NULL);
 	php_http_client_class_entry->create_object = php_http_client_object_new;
-	zend_class_implements(php_http_client_class_entry, 2, spl_ce_SplSubject, spl_ce_Countable);
+	zend_class_implements(php_http_client_class_entry, 2, spl_ce_SplSubject, zend_ce_countable);
 	memcpy(&php_http_client_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
 	php_http_client_object_handlers.offset = XtOffsetOf(php_http_client_object_t, zo);
 	php_http_client_object_handlers.free_obj = php_http_client_object_free;
 	php_http_client_object_handlers.clone_obj = NULL;
+	php_http_client_object_handlers.get_gc = php_http_client_object_get_gc;
 	zend_declare_property_null(php_http_client_class_entry, ZEND_STRL("observers"), ZEND_ACC_PRIVATE);
 	zend_declare_property_null(php_http_client_class_entry, ZEND_STRL("options"), ZEND_ACC_PROTECTED);
 	zend_declare_property_null(php_http_client_class_entry, ZEND_STRL("history"), ZEND_ACC_PROTECTED);
 	zend_declare_property_bool(php_http_client_class_entry, ZEND_STRL("recordHistory"), 0, ZEND_ACC_PUBLIC);
 
+	zend_declare_class_constant_long(php_http_client_class_entry, ZEND_STRL("DEBUG_INFO"), PHP_HTTP_CLIENT_DEBUG_INFO);
+	zend_declare_class_constant_long(php_http_client_class_entry, ZEND_STRL("DEBUG_IN"), PHP_HTTP_CLIENT_DEBUG_IN);
+	zend_declare_class_constant_long(php_http_client_class_entry, ZEND_STRL("DEBUG_OUT"), PHP_HTTP_CLIENT_DEBUG_OUT);
+	zend_declare_class_constant_long(php_http_client_class_entry, ZEND_STRL("DEBUG_HEADER"), PHP_HTTP_CLIENT_DEBUG_HEADER);
+	zend_declare_class_constant_long(php_http_client_class_entry, ZEND_STRL("DEBUG_BODY"), PHP_HTTP_CLIENT_DEBUG_BODY);
+	zend_declare_class_constant_long(php_http_client_class_entry, ZEND_STRL("DEBUG_SSL"), PHP_HTTP_CLIENT_DEBUG_SSL);
+
 	zend_hash_init(&php_http_client_drivers, 2, NULL, php_http_client_driver_hash_dtor, 1);
 
 	return SUCCESS;
