@@ -38,7 +38,7 @@ static inline void php_http_querystring_set(zval *instance, zval *params, int fl
 	array_init(&qa);
 
 	if (flags & QS_MERGE) {
-		zval old_tmp, *old = zend_read_property(php_http_querystring_class_entry, instance, ZEND_STRL("queryArray"), 0, &old_tmp);
+		zval old_tmp, *old = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(instance), ZEND_STRL("queryArray"), 0, &old_tmp);
 
 		ZVAL_DEREF(old);
 		if (Z_TYPE_P(old) == IS_ARRAY) {
@@ -47,13 +47,13 @@ static inline void php_http_querystring_set(zval *instance, zval *params, int fl
 	}
 
 	php_http_querystring_update(&qa, params, NULL);
-	zend_update_property(php_http_querystring_class_entry, instance, ZEND_STRL("queryArray"), &qa);
+	zend_update_property(php_http_querystring_class_entry, Z_OBJ_P(instance), ZEND_STRL("queryArray"), &qa);
 	zval_ptr_dtor(&qa);
 }
 
 static inline void php_http_querystring_str(zval *instance, zval *return_value)
 {
-	zval qa_tmp, *qa = zend_read_property(php_http_querystring_class_entry, instance, ZEND_STRL("queryArray"), 0, &qa_tmp);
+	zval qa_tmp, *qa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(instance), ZEND_STRL("queryArray"), 0, &qa_tmp);
 
 	ZVAL_DEREF(qa);
 	if (Z_TYPE_P(qa) == IS_ARRAY) {
@@ -65,7 +65,7 @@ static inline void php_http_querystring_str(zval *instance, zval *return_value)
 
 static inline void php_http_querystring_get(zval *instance, int type, char *name, uint32_t name_len, zval *defval, zend_bool del, zval *return_value)
 {
-	zval *arrval, qarray_tmp, *qarray = zend_read_property(php_http_querystring_class_entry, instance, ZEND_STRL("queryArray"), 0, &qarray_tmp);
+	zval *arrval, qarray_tmp, *qarray = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(instance), ZEND_STRL("queryArray"), 0, &qarray_tmp);
 
 	ZVAL_DEREF(qarray);
 	if ((Z_TYPE_P(qarray) == IS_ARRAY) && (arrval = zend_symtable_str_find(Z_ARRVAL_P(qarray), name, name_len))) {
@@ -98,7 +98,7 @@ ZEND_RESULT_CODE php_http_querystring_xlate(zval *dst, zval *src, const char *ie
 	zval *entry;
 	zend_string *xkey, *xstr;
 	php_http_arrkey_t key;
-	
+
 	ZEND_HASH_FOREACH_KEY_VAL(Z_ARRVAL_P(src), key.h, key.key, entry)
 	{
 		if (key.key) {
@@ -107,7 +107,7 @@ ZEND_RESULT_CODE php_http_querystring_xlate(zval *dst, zval *src, const char *ie
 				return FAILURE;
 			}
 		}
-		
+
 		if (Z_TYPE_P(entry) == IS_STRING) {
 			if (PHP_ICONV_ERR_SUCCESS != php_iconv_string(Z_STRVAL_P(entry), Z_STRLEN_P(entry), &xstr, oe, ie)) {
 				if (key.key) {
@@ -123,7 +123,7 @@ ZEND_RESULT_CODE php_http_querystring_xlate(zval *dst, zval *src, const char *ie
 			}
 		} else if (Z_TYPE_P(entry) == IS_ARRAY) {
 			zval subarray;
-			
+
 			array_init(&subarray);
 			if (key.key) {
 				add_assoc_zval_ex(dst, xkey->val, xkey->len, &subarray);
@@ -137,7 +137,7 @@ ZEND_RESULT_CODE php_http_querystring_xlate(zval *dst, zval *src, const char *ie
 				return FAILURE;
 			}
 		}
-		
+
 		if (key.key) {
 			zend_string_release(xkey);
 		}
@@ -254,7 +254,7 @@ ZEND_RESULT_CODE php_http_querystring_update(zval *qarray, zval *params, zval *o
 
 		/* squeeze the hash out of the zval */
 		if (Z_TYPE_P(params) == IS_OBJECT && instanceof_function(Z_OBJCE_P(params), php_http_querystring_class_entry)) {
-			zval qa_tmp, *qa = zend_read_property(php_http_querystring_class_entry, params, ZEND_STRL("queryArray"), 0, &qa_tmp);
+			zval qa_tmp, *qa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(params), ZEND_STRL("queryArray"), 0, &qa_tmp);
 
 			ZVAL_DEREF(qa);
 			convert_to_array(qa);
@@ -357,7 +357,7 @@ PHP_METHOD(HttpQueryString, __construct)
 {
 	zval *params = NULL;
 	zend_error_handling zeh;
-	
+
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "|z", &params), invalid_arg, return);
 
 	zend_replace_error_handling(EH_THROW, php_http_get_exception_bad_querystring_class_entry(), &zeh);
@@ -378,13 +378,12 @@ PHP_METHOD(HttpQueryString, getGlobalInstance)
 	if (Z_TYPE_P(instance) == IS_OBJECT) {
 		RETVAL_ZVAL(instance, 1, 0);
 	} else if ((_GET = php_http_env_get_superglobal(ZEND_STRL("_GET")))) {
-		zval tmp, *qa;
+		zval *qa;
+		zend_string *qa_str = zend_string_init(ZEND_STRL("queryArray"), 0);
 
 		ZVAL_OBJ(return_value, php_http_querystring_object_new(php_http_querystring_class_entry));
-
-		ZVAL_STRING(&tmp, "queryArray");
-		qa = Z_OBJ_HT_P(return_value)->get_property_ptr_ptr(return_value, &tmp, BP_VAR_RW, NULL);
-		zval_ptr_dtor(&tmp);
+		qa = Z_OBJ_HT_P(return_value)->get_property_ptr_ptr(Z_OBJ_P(return_value), qa_str, BP_VAR_RW, NULL);
+		zend_string_release(qa_str);
 
 		ZVAL_NEW_REF(_GET, _GET);
 		ZVAL_COPY(qa, _GET);
@@ -396,7 +395,7 @@ PHP_METHOD(HttpQueryString, getGlobalInstance)
 
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_getIterator, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_RETURN_OBJ_INFO_EX(ai_HttpQueryString_getIterator, 0, 0, Traversable, 0)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, getIterator)
 {
@@ -404,10 +403,10 @@ PHP_METHOD(HttpQueryString, getIterator)
 
 	php_http_expect(SUCCESS == zend_parse_parameters_none(), invalid_arg, return);
 
-	qa = zend_read_property(php_http_querystring_class_entry, getThis(), ZEND_STRL("queryArray"), 0, &qa_tmp);
+	qa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("queryArray"), 0, &qa_tmp);
 
 	object_init_ex(return_value, spl_ce_RecursiveArrayIterator);
-	zend_call_method_with_1_params(return_value, spl_ce_RecursiveArrayIterator, NULL, "__construct", NULL, qa);
+	zend_call_method_with_1_params(Z_OBJ_P(return_value), spl_ce_RecursiveArrayIterator, NULL, "__construct", NULL, qa);
 }
 
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_toString, 0, 0, 0)
@@ -430,7 +429,7 @@ PHP_METHOD(HttpQueryString, toArray)
 		return;
 	}
 
-	zqa = zend_read_property(php_http_querystring_class_entry, getThis(), ZEND_STRL("queryArray"), 0, &zqa_tmp);
+	zqa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("queryArray"), 0, &zqa_tmp);
 	RETURN_ZVAL(zqa, 1, 0);
 }
 
@@ -447,7 +446,7 @@ PHP_METHOD(HttpQueryString, get)
 	zend_long type = 0;
 	zend_bool del = 0;
 	zval *ztype = NULL, *defval = NULL;
-	
+
 	if (SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "|szzb", &name_str, &name_len, &ztype, &defval, &del)) {
 		if (name_str && name_len) {
 			if (ztype) {
@@ -455,7 +454,7 @@ PHP_METHOD(HttpQueryString, get)
 					type = Z_LVAL_P(ztype);
 				} else if(Z_TYPE_P(ztype) == IS_STRING) {
 					switch (Z_STRVAL_P(ztype)[0]) {
-						case 'B': 
+						case 'B':
 						case 'b':	type = PHP_HTTP_QUERYSTRING_TYPE_BOOL;		break;
 						case 'L':
 						case 'l':
@@ -464,7 +463,7 @@ PHP_METHOD(HttpQueryString, get)
 						case 'd':
 						case 'D':
 						case 'F':
-						case 'f':	type = PHP_HTTP_QUERYSTRING_TYPE_FLOAT;		break;	
+						case 'f':	type = PHP_HTTP_QUERYSTRING_TYPE_FLOAT;		break;
 						case 'S':
 						case 's':	type = PHP_HTTP_QUERYSTRING_TYPE_STRING;	break;
 						case 'A':
@@ -487,11 +486,11 @@ ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, set)
 {
 	zval *params;
-	
+
 	if (SUCCESS != zend_parse_parameters(ZEND_NUM_ARGS(), "z", &params)) {
 		return;
 	}
-	
+
 	php_http_querystring_set(getThis(), params, QS_MERGE);
 	RETVAL_ZVAL(getThis(), 1, 0);
 }
@@ -505,11 +504,11 @@ PHP_METHOD(HttpQueryString, mod)
 	zend_error_handling zeh;
 
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "z", &params), invalid_arg, return);
-	
+
 	zend_replace_error_handling(EH_THROW, php_http_get_exception_bad_querystring_class_entry(), &zeh);
-	ZVAL_OBJ(return_value, Z_OBJ_HT_P(instance)->clone_obj(instance));
+	ZVAL_OBJ(return_value, Z_OBJ_HT_P(instance)->clone_obj(Z_OBJ_P(instance)));
 	/* make sure we do not inherit the reference to _GET */
-	SEPARATE_ZVAL(zend_read_property(Z_OBJCE_P(return_value), return_value, ZEND_STRL("queryArray"), 0, &qa_tmp));
+	SEPARATE_ZVAL(zend_read_property(Z_OBJCE_P(return_value), Z_OBJ_P(return_value), ZEND_STRL("queryArray"), 0, &qa_tmp));
 	php_http_querystring_set(return_value, params, QS_MERGE);
 	zend_restore_error_handling(&zeh);
 }
@@ -551,7 +550,7 @@ PHP_METHOD(HttpQueryString, xlate)
 	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "ss", &ie, &ie_len, &oe, &oe_len), invalid_arg, return);
 
 	array_init(&na);
-	qa = zend_read_property(php_http_querystring_class_entry, getThis(), ZEND_STRL("queryArray"), 0, &qa_tmp);
+	qa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("queryArray"), 0, &qa_tmp);
 	ZVAL_DEREF(qa);
 	convert_to_array(qa);
 
@@ -562,11 +561,34 @@ PHP_METHOD(HttpQueryString, xlate)
 
 	php_http_querystring_set(getThis(), &na, 0);
 	RETVAL_ZVAL(getThis(), 1, 0);
-	
+
 	zval_ptr_dtor(&na);
 }
 #endif /* HAVE_ICONV */
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpQueryString___serialize, 0, 0, IS_ARRAY, 0)
+ZEND_END_ARG_INFO();
+PHP_METHOD(HttpQueryString, __serialize)
+{
+	zval *zqa, zqa_tmp;
+
+	zend_parse_parameters_none();
+
+	zqa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("queryArray"), 0, &zqa_tmp);
+	RETURN_ZVAL(zqa, 1, 0);
+}
+
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpQueryString___unserialize, 0, 1, IS_VOID, 0)
+	ZEND_ARG_TYPE_INFO(0, data, IS_ARRAY, 0)
+ZEND_END_ARG_INFO();
+PHP_METHOD(HttpQueryString, __unserialize)
+{
+	zval *qa;
+
+	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "a", &qa), invalid_arg, return);
+	php_http_querystring_set(getThis(), qa, 0);
+}
+
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_serialize, 0, 0, 0)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, serialize)
@@ -583,7 +605,7 @@ ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, unserialize)
 {
 	zval *serialized;
-	
+
 	if (SUCCESS != zend_parse_parameters(ZEND_NUM_ARGS(), "z", &serialized)) {
 		return;
 	}
@@ -595,8 +617,8 @@ PHP_METHOD(HttpQueryString, unserialize)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_offsetGet, 0, 0, 1)
-	ZEND_ARG_INFO(0, offset)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpQueryString_offsetGet, 0, 1, IS_MIXED, 1)
+	ZEND_ARG_INFO(0, name)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, offsetGet)
 {
@@ -606,8 +628,8 @@ PHP_METHOD(HttpQueryString, offsetGet)
 	if (SUCCESS != zend_parse_parameters(ZEND_NUM_ARGS(), "S", &offset)) {
 		return;
 	}
-	
-	qa = zend_read_property(php_http_querystring_class_entry, getThis(), ZEND_STRL("queryArray"), 0, &qa_tmp);
+
+	qa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("queryArray"), 0, &qa_tmp);
 	ZVAL_DEREF(qa);
 
 	if (Z_TYPE_P(qa) == IS_ARRAY) {
@@ -617,15 +639,15 @@ PHP_METHOD(HttpQueryString, offsetGet)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_offsetSet, 0, 0, 2)
-	ZEND_ARG_INFO(0, offset)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpQueryString_offsetSet, 0, 2, IS_VOID, 0)
+	ZEND_ARG_INFO(0, name)
 	ZEND_ARG_INFO(0, value)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, offsetSet)
 {
 	zend_string *offset;
 	zval *value, param, znull;
-	
+
 	if (SUCCESS != zend_parse_parameters(ZEND_NUM_ARGS(), "Sz", &offset, &value)) {
 		return;
 	}
@@ -642,8 +664,8 @@ PHP_METHOD(HttpQueryString, offsetSet)
 	zval_ptr_dtor(&param);
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_offsetExists, 0, 0, 1)
-	ZEND_ARG_INFO(0, offset)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpQueryString_offsetExists, 0, 1, _IS_BOOL, 0)
+	ZEND_ARG_INFO(0, name)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, offsetExists)
 {
@@ -653,8 +675,8 @@ PHP_METHOD(HttpQueryString, offsetExists)
 	if (SUCCESS != zend_parse_parameters(ZEND_NUM_ARGS(), "S", &offset)) {
 		return;
 	}
-	
-	qa = zend_read_property(php_http_querystring_class_entry, getThis(), ZEND_STRL("queryArray"), 0, &qa_tmp);
+
+	qa = zend_read_property(php_http_querystring_class_entry, Z_OBJ_P(ZEND_THIS), ZEND_STRL("queryArray"), 0, &qa_tmp);
 	ZVAL_DEREF(qa);
 
 	if (Z_TYPE_P(qa) == IS_ARRAY) {
@@ -665,14 +687,14 @@ PHP_METHOD(HttpQueryString, offsetExists)
 	RETURN_FALSE;
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpQueryString_offsetUnset, 0, 0, 1)
-	ZEND_ARG_INFO(0, offset)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpQueryString_offsetUnset, 0, 1, IS_VOID, 0)
+	ZEND_ARG_INFO(0, name)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpQueryString, offsetUnset)
 {
 	zend_string *offset;
 	zval param, znull;
-	
+
 	if (SUCCESS != zend_parse_parameters(ZEND_NUM_ARGS(), "S", &offset)) {
 		return;
 	}
@@ -684,12 +706,15 @@ PHP_METHOD(HttpQueryString, offsetUnset)
 	zval_ptr_dtor(&param);
 }
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpQueryString___toString, 0, 0, IS_STRING, 0)
+ZEND_END_ARG_INFO();
+
 static zend_function_entry php_http_querystring_methods[] = {
 	PHP_ME(HttpQueryString, __construct, ai_HttpQueryString___construct, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
 
 	PHP_ME(HttpQueryString, toArray, ai_HttpQueryString_toArray, ZEND_ACC_PUBLIC)
 	PHP_ME(HttpQueryString, toString, ai_HttpQueryString_toString, ZEND_ACC_PUBLIC)
-	ZEND_MALIAS(HttpQueryString, __toString, toString, ai_HttpQueryString_toString, ZEND_ACC_PUBLIC)
+	ZEND_MALIAS(HttpQueryString, __toString, toString, ai_HttpQueryString___toString, ZEND_ACC_PUBLIC)
 
 	PHP_ME(HttpQueryString, get, ai_HttpQueryString_get, ZEND_ACC_PUBLIC)
 	PHP_ME(HttpQueryString, set, ai_HttpQueryString_set, ZEND_ACC_PUBLIC)
@@ -712,6 +737,8 @@ static zend_function_entry php_http_querystring_methods[] = {
 	/* Implements Serializable */
 	PHP_ME(HttpQueryString, serialize, ai_HttpQueryString_serialize, ZEND_ACC_PUBLIC)
 	PHP_ME(HttpQueryString, unserialize, ai_HttpQueryString_unserialize, ZEND_ACC_PUBLIC)
+	PHP_ME(HttpQueryString, __serialize, ai_HttpQueryString___serialize, ZEND_ACC_PUBLIC)
+	PHP_ME(HttpQueryString, __unserialize, ai_HttpQueryString___unserialize, ZEND_ACC_PUBLIC)
 
 	/* Implements ArrayAccess */
 	PHP_ME(HttpQueryString, offsetGet, ai_HttpQueryString_offsetGet, ZEND_ACC_PUBLIC)
